@@ -1,5 +1,7 @@
 "elo" <- function(x, status=NULL, init=2200, gamma=0, kfac=27, history=FALSE, sort=TRUE, ...)
 {
+  if(!is.data.frame(x)) x <- as.data.frame(x)
+  if(!is.data.frame(status) && !is.null(status)) status <- as.data.frame(status)
   if(length(init) != 1) stop("the length of 'init' must be one")
   if(ncol(x) != 4) stop("'x' must have four variables")
   if(nrow(x) == 0) {
@@ -23,7 +25,6 @@
   np <- length(play)
   x$White <- match(x$White, play)
   x$Black <- match(x$Black, play)
-
   if(!is.null(status)) {
     npadd <- play[!(play %in% status$Player)]
 	zv <- rep(0, length(npadd))
@@ -124,6 +125,8 @@
 "fide" <- function(x, status=NULL, init=2200, gamma=0, kfac=kfide,
   history=FALSE, sort=TRUE, ...)
 {  
+  if(!is.data.frame(x)) x <- as.data.frame(x)
+  if(!is.data.frame(status) && !is.null(status)) status <- as.data.frame(status)
   if(length(init) != 1) stop("the length of 'init' must be one")
   if(ncol(x) != 4) stop("'x' must have four variables")
   if(nrow(x) == 0) {
@@ -267,9 +270,13 @@
   lout
 }
 
-"glicko" <- function(x, status=NULL, init=c(2200,300), gamma=0, cval=15, history=FALSE, sort=TRUE, ...)
+"glicko" <- function(x, status=NULL, init=c(2200,300), gamma=0, cval=15, history=FALSE, sort=TRUE, rdmax = 350, ...)
 { 
+  if(!is.data.frame(x)) x <- as.data.frame(x)
+  if(!is.data.frame(status) && !is.null(status)) status <- as.data.frame(status)
   if(length(init) != 2) stop("the length of 'init' must be two")
+  if(init[2] <= 0) stop("initial deviation must be positive")
+  if(init[2] > rdmax) stop("initial deviation cannot be greater than rdmax")
   if(ncol(x) != 4) stop("'x' must have four variables")
   if(nrow(x) == 0) {
     if(is.null(status)) stop("'x' is empty and 'status' is NULL")
@@ -322,8 +329,6 @@
     names(rinit) <- names(dinit) <- names(ngames) <- names(nlag) <- play
   }
  
-  if(!all(names(rinit) == names(ngames)))
-    stop("names of ratings and deviations are different")
   if(!all(names(rinit) == names(ngames)))
     stop("names of ratings and ngames are different")
   if(!all(play %in% names(rinit))) 
@@ -347,7 +352,7 @@
   nloss <- nloss[curplay]
   nlag <- nlag[curplay]
 
-  qv <- log(10)/400; qv2 <- qv^2; qip3 <- 3*(qv/pi)^2 
+  qv <- log(10)/400; qip3 <- 3*(qv/pi)^2 
   gammas <- split(gammas, x$Month)
   x <- split(x, x$Month)
   players <- lapply(x, function(y) unique(c(y$White, y$Black)))
@@ -361,7 +366,7 @@
     nr <- nrow(traini)
     playi <- players[[i]]
 
-    cdevs[playi] <- pmin(cdevs[playi] + (nlag[playi]+1)*(cval^2), 122500)
+    cdevs[playi] <- pmin(cdevs[playi] + (nlag[playi]+1)*(cval^2), rdmax * rdmax)
     gdevs <- 1/sqrt(1 + qip3*cdevs) 
     ngamesi <- tabulate(c(traini$White,traini$Black), np)
     dscore <- .C("glicko_c",
@@ -404,10 +409,193 @@
   lout
 }
 
+"glicko2" <- function(x, status=NULL, init=c(2200,300,0.15), gamma=0, tau=1.2, history=FALSE, sort=TRUE, rdmax = 350, ...)
+{ 
+  if(!is.data.frame(x)) x <- as.data.frame(x)
+  if(!is.data.frame(status) && !is.null(status)) status <- as.data.frame(status)
+  qv <- log(10)/400
+  if(length(init) != 3) stop("the length of 'init' must be three")
+  if(init[2] <= 0) stop("initial deviation must be positive")
+  if(init[2] > rdmax) stop("initial deviation cannot be greater than rdmax")
+  if(init[3] <= 0) stop("initial volatility must be positive")
+  if(init[3] > qv*rdmax) stop("initial volatility cannot be greater than log(10)*rdmax/400")
+  if(ncol(x) != 4) stop("'x' must have four variables")
+  if(nrow(x) == 0) {
+    if(is.null(status)) stop("'x' is empty and 'status' is NULL")
+    lout <- list(ratings=status, history=NULL, gamma=gamma, tau=tau, type = "Glicko-2")
+    class(lout) <- "rating"
+    return(lout)
+  }
+  gammas <- rep(gamma, length.out = nrow(x)) 
+  names(x) <- c("Month","White","Black","Score")
+  if(!is.numeric(x$Month)) 
+    stop("Time period must be numeric")
+  if(!is.numeric(x$White) && !is.character(x$White))
+    stop("Player identifiers must be numeric or character")
+  if(!is.numeric(x$Black) && !is.character(x$Black))
+    stop("Player identifiers must be numeric or character")	
+  if(!is.numeric(x$Score) || any(x$Score > 1) || any(x$Score < 0))
+    stop("Game scores must be in the interval [0,1]")
+  
+  play <- sort(unique(c(x$White,x$Black)))
+  np <- length(play)
+  x$White <- match(x$White, play)
+  x$Black <- match(x$Black, play)
+  
+  if(!is.null(status)) {
+    npadd <- play[!(play %in% status$Player)]
+    zv <- rep(0, length(npadd))
+    npstatus <- data.frame(Player = npadd, Rating = rep(init[1],length(npadd)), 
+                           Deviation = rep(init[2],length(npadd)), Volatility = rep(init[3],length(npadd)), 
+                           Games = zv, Win = zv, Draw = zv, Loss = zv, Lag = zv)
+    if(!("Games" %in% names(status))) status <- cbind(status, Games = 0)
+    if(!("Win" %in% names(status))) status <- cbind(status, Win = 0)
+    if(!("Draw" %in% names(status))) status <- cbind(status, Draw = 0)
+    if(!("Loss" %in% names(status))) status <- cbind(status, Loss = 0)
+    if(!("Lag" %in% names(status))) status <- cbind(status, Lag = 0)
+    status <- rbind(status[,c("Player","Rating","Deviation","Volatility","Games","Win","Draw","Loss","Lag")], npstatus)
+    rinit <- status[[2]]
+    dinit <- status[[3]]
+    vinit <- status[[4]]
+    ngames <- status[[5]]
+    nwin <- status[[6]]
+    ndraw <- status[[7]]
+    nloss <- status[[8]]
+    nlag <- status[[9]]
+    names(rinit) <- names(dinit) <- names(vinit) <- names(ngames) <- status$Player
+  }
+  else {
+    rinit <- rep(init[1], length.out=np)
+    dinit <- rep(init[2], length.out=np)
+    vinit <- rep(init[3], length.out=np)
+    ngames <- nwin <- ndraw <- nloss <- rep(0, length.out=np)
+    nlag <- rep(0,np)
+    names(rinit) <- names(dinit) <- names(vinit) <- names(ngames) <- names(nlag) <- play
+  }
+  
+  if(!all(names(rinit) == names(ngames)))
+    stop("names of ratings and ngames are different")
+  if(!all(play %in% names(rinit))) 
+    stop("Payers in data are not within current status")
+  
+  # conversion to Glicko-2 scale
+  rinit <- qv*(rinit - 1500)
+  gammas <- qv*gammas
+  dinit <- qv*dinit
+  
+  nm <- length(unique(x$Month))
+  curplay <- match(play, names(rinit))
+  orats <- rinit[-curplay] 
+  odevs <- dinit[-curplay]^2
+  ovols <- vinit[-curplay]
+  ongames <- ngames[-curplay]
+  onwin <- nwin[-curplay]
+  ondraw <- ndraw[-curplay]
+  onloss <- nloss[-curplay]
+  olag <- nlag[-curplay]
+  olag[ongames != 0] <- olag[ongames != 0] + nm
+  crats <- rinit[curplay] 
+  cdevs <- dinit[curplay]^2
+  cvols <- vinit[curplay]
+  ngames <- ngames[curplay] 
+  nwin <- nwin[curplay]
+  ndraw <- ndraw[curplay]
+  nloss <- nloss[curplay]
+  nlag <- nlag[curplay]
+  
+  gammas <- split(gammas, x$Month)
+  x <- split(x, x$Month)
+  players <- lapply(x, function(y) unique(c(y$White, y$Black)))
+  if(history) {
+    histry <- array(NA, dim=c(np,nm,5), dimnames=list(play,1:nm,c("Rating","Deviation","Volatility","Games","Lag")))
+  }
+  
+  tau2 <- tau * tau
+  for(i in 1:nm) {
+    traini <- x[[i]]
+    gammai <- gammas[[i]]
+    nr <- nrow(traini)
+    playi <- players[[i]]
+    
+    # nlag[payi]*(cvols[playi]^2) in Glicko-2, (nlag[payi]+1)*(cval^2) in Glicko
+    rdmax2 <- qv * qv * rdmax * rdmax
+    cdevs[playi] <- pmin(cdevs[playi] + (nlag[playi])*(cvols[playi]^2), rdmax2)
+    qip3 <- 3*(1/pi)^2 
+    gdevs <- 1/sqrt(1 + qip3*cdevs) 
+    ngamesi <- tabulate(c(traini$White,traini$Black), np)
+    dscore <- .C("glicko2_c",
+                 as.integer(np), as.integer(nr), as.integer(traini$White-1), as.integer(traini$Black-1),
+                 as.double(traini$Score), as.double(crats), as.double(gdevs), as.double(gammai),
+                 dscore = double(2*np))$dscore
+    vval <- dscore[(np+1):(2*np)]; dscore <- dscore[1:np]
+    
+    if(tau > 0) {
+      nllh <- function(z, pz, cdev, vvi, dsco) {
+        denom <- cdev + exp(z) + vvi
+        delta <- vvi * dsco
+        (z - pz) * (z - pz) / tau2 + log(denom) + delta*delta/denom
+      }
+      for(k in seq_along(playi)) {
+        prv <- 2*log(cvols[playi[k]])
+        oval <- optimize(nllh, lower = prv-4*tau, upper = prv+4*tau, pz = prv, 
+          cdev = cdevs[playi[k]], vvi = 1/vval[playi[k]], dsco = dscore[playi[k]])$minimum
+        cvols[playi[k]] <- min(exp(oval/2), qv * rdmax) 
+      }
+    }
+    
+    cdevs[playi] <- cdevs[playi] + cvols[playi]^2
+    cdevs <- pmin(1/(1/cdevs + vval), rdmax2)
+    crats <- crats + cdevs * dscore
+    
+    trainiplw <- c(traini$White[traini$Score==1],traini$Black[traini$Score==0])
+    trainipld <- c(traini$White[traini$Score==0.5],traini$Black[traini$Score==0.5])
+    trainipll <- c(traini$White[traini$Score==0],traini$Black[traini$Score==1])
+    ngames <- ngames + ngamesi
+    nwin <- nwin + tabulate(trainiplw, np)
+    ndraw <- ndraw + tabulate(trainipld, np)
+    nloss <- nloss + tabulate(trainipll, np)
+    nlag[ngames!=0] <- nlag[ngames!=0] + 1
+    nlag[playi] <- 0
+    
+    if(history) {
+      histry[,i,1] <- crats
+      histry[,i,2] <- sqrt(cdevs)
+      histry[,i,3] <- cvols
+      histry[,i,4] <- ngames
+      histry[,i,5] <- nlag
+    }
+  }
+  if(!history) histry <- NULL
+  player <- suppressWarnings(as.numeric(names(c(crats,orats))))
+  if (any(is.na(player))) player <- names(c(crats,orats))
+  dfout <- data.frame(Player=player, Rating=c(crats,orats), Deviation=sqrt(c(cdevs,odevs)), Volatility=c(cvols,ovols), 
+                      Games=c(ngames,ongames), Win=c(nwin,onwin), Draw=c(ndraw,ondraw), Loss=c(nloss,onloss), 
+                      Lag=c(nlag,olag),
+                      stringsAsFactors = FALSE)
+  if(sort) dfout <- dfout[order(dfout$Rating,decreasing=TRUE),] else dfout <- dfout[order(dfout$Player),]
+  row.names(dfout) <- 1:nrow(dfout)
+  
+  # conversion from Glicko-2 scale
+  dfout$Rating <- dfout$Rating/qv + 1500
+  dfout$Deviation <- dfout$Deviation/qv
+  if(history) {
+    histry[,,1] <- histry[,,1]/qv + 1500
+    histry[,,2] <- histry[,,2]/qv
+  }
+  
+  lout <- list(ratings=dfout, history=histry, gamma=gamma, tau=tau, type = "Glicko-2")
+  class(lout) <- "rating"
+  lout
+}
+
 "steph" <- function(x, status=NULL, init=c(2200,300), gamma=0, cval=10, hval=10, 
-  bval=0, lambda = 2, history=FALSE, sort=TRUE, ...)
+  bval=0, lambda = 2, history=FALSE, sort=TRUE, rdmax = 350, ...)
 {
+  if(!is.data.frame(x)) x <- as.data.frame(x)
+  if(!is.data.frame(status) && !is.null(status)) status <- as.data.frame(status)
   if(length(init) != 2) stop("the length of 'init' must be two")
+  if(init[2] <= 0) stop("initial deviation must be positive")
+  if(init[2] > rdmax) stop("initial deviation cannot be greater than rdmax")
   if(ncol(x) != 4) stop("'x' must have four variables")
   if(nrow(x) == 0) {
     if(is.null(status)) stop("'x' is empty and 'status' is NULL")
@@ -461,8 +649,6 @@
     names(rinit) <- names(dinit) <- names(ngames) <- names(nlag) <- play
   }
 
-  if(!all(names(rinit) == names(ngames)))
-    stop("names of ratings and deviations are different")
   if(!all(names(rinit) == names(ngames)))
     stop("names of ratings and ngames are different")
   if(!all(play %in% names(rinit))) 
@@ -486,7 +672,7 @@
   nloss <- nloss[curplay]
   nlag <- nlag[curplay]
 
-  qv <- log(10)/400; qv2 <- qv^2; qip3 <- 3*(qv/pi)^2
+  qv <- log(10)/400; qip3 <- 3*(qv/pi)^2
   gammas <- split(gammas, x$Month)  
   x <- split(x, x$Month)
   players <- lapply(x, function(y) unique(c(y$White, y$Black)))
@@ -500,7 +686,7 @@
     nr <- nrow(traini)
     playi <- players[[i]]
 
-    cdevs[playi] <- pmin(cdevs[playi] + (nlag[playi]+1)*(cval^2), 122500)
+    cdevs[playi] <- pmin(cdevs[playi] + (nlag[playi]+1)*(cval^2), rdmax * rdmax)
     gdevs <- 1/sqrt(1 + qip3*cdevs) 
     ngamesi <- tabulate(c(traini$White,traini$Black), np)
     
@@ -620,15 +806,17 @@
   kfac
 }
 
-"print.rating" <-  function(x, digits = 0, ...) 
+"print.rating" <-  function(x, cols = 1:ncol(x$ratings), digits = 0, ...) 
 {
     rdf <- x$ratings
     rdf$Rating <- round(rdf$Rating, digits)
-    if(x$type == "Glicko" || x$type == "Stephenson") 
-	  rdf$Deviation <- round(rdf$Deviation, digits+2)
+    if(x$type == "Glicko" || x$type == "Glicko-2" || x$type == "Stephenson") 
+	    rdf$Deviation <- round(rdf$Deviation, digits+2)
+    if(x$type == "Glicko-2") 
+      rdf$Volatility <- round(rdf$Volatility, digits+4)
     np <- nrow(rdf); ng <- round(sum(rdf$Games)/2)
     cat(paste("\n",x$type," Ratings For ",np," Players Playing ",ng," Games\n\n", sep=""))
-    print(rdf[1:(min(1000,np)),])
+    print(rdf[1:(min(1000,np)),cols,drop=FALSE])
     if(np > 1000) cat("\nOutput Tructated To First 1000 Players \n")
     cat("\n")
     invisible(0)
@@ -645,30 +833,40 @@
 "predict.rating" <- function(object, newdata, tng=15, trat=NULL, gamma=30, 
   thresh, ...)
 {
-  if(nrow(newdata) == 0) stop("'newdata' must have non-zero rows")
+  if(missing(newdata) || nrow(newdata) == 0) 
+    stop("'newdata' must be non-missing and have non-zero rows")
   obj <- object$ratings
   wmat <- match(newdata[[2]], obj$Player)
   bmat <- match(newdata[[3]], obj$Player)
+  qv <- log(10)/400; qip3 <- 3*(qv/pi)^2
+  
+  if(!is.null(trat)) {
+    if(object$type == "Elo" && length(trat) != 1)
+      stop("'trat' must be vector of length one")
+    if(object$type != "Elo" && length(trat) != 2)
+      stop("'trat' must be vector of length two")
+  }
 
   if(!is.null(trat)) obj$Rating[obj$Games < tng] <- trat[1]
   else is.na(obj$Rating[obj$Games < tng]) <- TRUE
-  if(!is.null(trat)) obj$Deviation[obj$Games < tng] <- trat[2]
-  else is.na(obj$Deviation[obj$Games < tng]) <- TRUE
+  if(object$type != "Elo") {
+    if(!is.null(trat)) obj$Deviation[obj$Games < tng] <- trat[2]
+    else is.na(obj$Deviation[obj$Games < tng]) <- TRUE
+  }
+  
   wrat <- obj$Rating[wmat]; brat <- obj$Rating[bmat]
   if(!is.null(trat)) wrat[is.na(wrat)] <- trat[1]
   if(!is.null(trat)) brat[is.na(brat)] <- trat[1] 
-
-  qv <- log(10)/400; qip3 <- 3*(qv/pi)^2
-  if(object$type == "Glicko" || object$type == "Stephenson") {
-    if(!is.null(trat) && length(trat) != 2) 
-	  stop("'trat' must be vector of length two")
+  if(object$type != "Elo") 
+  {
     wdev <- obj$Deviation[wmat]; bdev <- obj$Deviation[bmat]
     if(!is.null(trat)) wdev[is.na(wdev)] <- trat[2]
     if(!is.null(trat)) bdev[is.na(bdev)] <- trat[2]
   }
+  
   if(object$type == "Elo")
     preds <- 1/(1+10^((brat-wrat-gamma)/400))
-  if(object$type == "Glicko" || object$type == "Stephenson") {
+  if(object$type != "Elo") {
     vec <- 1/sqrt(1 + qip3*(wdev^2 + bdev^2))
     preds <- 1/(1+10^(vec * (brat-wrat-gamma)/400))
   }
